"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.enhanceTenantEntity = exports.hooks = exports.beforeFindAfterExpandIncludeAll = exports.beforeUpsert = exports.beforeBulkDestroy = exports.beforeDestroy = exports.beforeBulkUpdate = exports.beforeUpdate = exports.beforeBulkCreate = exports.beforeCreate = exports.injectQueryScopeWithTenantProperty = exports.injectInstanceWithTenantProperty = void 0;
const Sequelize = require("sequelize");
const sequelize_typescript_1 = require("sequelize-typescript");
const entity_1 = require("./entity");
const constants_1 = require("./constants");
const decorator_error_1 = require("../errors/decorator.error");
function tenancyService(target) {
    const service = (0, entity_1.getTenancyService)(target);
    if (!service) {
        throw new decorator_error_1.DecoratorError(`MTService hasn't been injected into model scope of "${target.name}"`);
    }
    return service;
}
function tenancyDisabledViaOptions(options, removeKey = true) {
    if ((options || {})[constants_1.DISABLE_TENANCY_OPTION]) {
        if (removeKey) {
            delete options[constants_1.DISABLE_TENANCY_OPTION];
        }
        return true;
    }
    return false;
}
function injectInstanceWithTenantProperty(attributes, options, target, tenancyOptions) {
    if (!(0, entity_1.isTenantEntity)(target) || tenancyDisabledViaOptions(options)) {
        return;
    }
    const service = tenancyService(target);
    if (!service.tenancyScope.enabled) {
        return;
    }
    if (!attributes[tenancyOptions.tenantField]) {
        attributes[tenancyOptions.tenantField] = service.tenancyScope.tenant;
    }
}
exports.injectInstanceWithTenantProperty = injectInstanceWithTenantProperty;
function injectQueryScopeWithTenantProperty(options, target, tenancyOptions) {
    if (!(0, entity_1.isTenantEntity)(target) || tenancyDisabledViaOptions(options)) {
        return;
    }
    const service = tenancyService(target);
    if (!service.tenancyScope.enabled) {
        return;
    }
    options = options || {};
    options.where = options.where || {};
    const tenantWhereClause = service.tenancyOptions.allowMissingTenant
        ? { [Sequelize.Op.in]: [service.tenancyScope.tenant, null] }
        : service.tenancyScope.tenant;
    if (!options.where[tenancyOptions.tenantField]) {
        options.where[tenancyOptions.tenantField] = tenantWhereClause;
    }
    for (const nestedOptions of options.include || []) {
        injectQueryScopeWithTenantProperty(nestedOptions, nestedOptions.model, tenancyOptions);
    }
}
exports.injectQueryScopeWithTenantProperty = injectQueryScopeWithTenantProperty;
function beforeCreate(target, tenancyOptions) {
    return (attributes, options) => {
        injectInstanceWithTenantProperty(attributes, options, target, tenancyOptions);
    };
}
exports.beforeCreate = beforeCreate;
function beforeBulkCreate(target, tenancyOptions) {
    return (items, options) => {
        for (const attributes of items) {
            injectInstanceWithTenantProperty(attributes, options, target, tenancyOptions);
        }
    };
}
exports.beforeBulkCreate = beforeBulkCreate;
function beforeUpdate(target, tenancyOptions) {
    return (attributes, options) => {
        injectInstanceWithTenantProperty(attributes, options, target, tenancyOptions);
    };
}
exports.beforeUpdate = beforeUpdate;
function beforeBulkUpdate(target, tenancyOptions) {
    return (options) => {
        injectQueryScopeWithTenantProperty(options, target, tenancyOptions);
    };
}
exports.beforeBulkUpdate = beforeBulkUpdate;
function beforeDestroy(target, tenancyOptions) {
    return (attributes, options) => {
        injectInstanceWithTenantProperty(attributes, options, target, tenancyOptions);
    };
}
exports.beforeDestroy = beforeDestroy;
function beforeBulkDestroy(target, tenancyOptions) {
    return (options) => {
        injectQueryScopeWithTenantProperty(options, target, tenancyOptions);
    };
}
exports.beforeBulkDestroy = beforeBulkDestroy;
function beforeUpsert(target, tenancyOptions) {
    return (attributes, options) => {
        injectInstanceWithTenantProperty(attributes, options, target, tenancyOptions);
        injectQueryScopeWithTenantProperty(options, target, tenancyOptions);
    };
}
exports.beforeUpsert = beforeUpsert;
function beforeFindAfterExpandIncludeAll(target, tenancyOptions) {
    return (options) => {
        injectQueryScopeWithTenantProperty(options, target, tenancyOptions);
    };
}
exports.beforeFindAfterExpandIncludeAll = beforeFindAfterExpandIncludeAll;
exports.hooks = [
    beforeCreate,
    beforeBulkCreate,
    beforeUpdate,
    beforeBulkUpdate,
    beforeUpsert,
    beforeDestroy,
    beforeBulkDestroy,
    beforeFindAfterExpandIncludeAll,
];
function enhanceTenantEntity(target, tenancyOptions) {
    for (const hook of exports.hooks) {
        const hookMethod = `${constants_1.HOOK_METHOD_PREFIX}$${hook.name}`;
        target[hookMethod] = hook(target, tenancyOptions);
        (0, sequelize_typescript_1.addHook)(target, hook.name, hookMethod);
    }
}
exports.enhanceTenantEntity = enhanceTenantEntity;
