"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.MtenantService = void 0;
const tslib_1 = require("tslib");
const common_1 = require("@nestjs/common");
const async_context_1 = require("@nestjs-steroids/async-context");
const mtenant_error_1 = require("./errors/mtenant.error");
const entity_1 = require("./decorators/entity");
const core_interface_1 = require("./interfaces/core.interface");
const http_transport_1 = require("./transports/http.transport");
const ioredis_1 = require("./storages/cache/ioredis");
const cached_storage_1 = require("./storages/cached.storage");
const sequelize_storage_1 = require("./storages/sequelize.storage");
const constants_1 = require("./constants");
let MtenantService = class MtenantService {
    constructor(options, asyncContext) {
        this.options = options;
        this.asyncContext = asyncContext;
        this.setup(options);
    }
    runWithinTenancyScope(context, handler) {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            this.asyncContext.register();
            this.asyncContext.set(constants_1.MT_SCOPE_KEY, {
                tenant: yield this.getTenant(context),
                enabled: true,
            });
            handler();
        });
    }
    setTenant(tenant, context) {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            if (!(yield this.isTenantAllowed(context || {}, tenant))) {
                throw new common_1.NotAcceptableException(`Tenant "${tenant}" not allowed`);
            }
            this.asyncContext.set(constants_1.MT_SCOPE_KEY, {
                tenant,
                enabled: true,
            });
            return this;
        });
    }
    getTenant(context) {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            const tenant = yield this.transport.extract(context);
            if (tenant && !(yield this.isTenantAllowed(context, tenant))) {
                throw new common_1.NotAcceptableException(`Tenant "${tenant}" not allowed`);
            }
            return tenant || this.options.defaultTenant;
        });
    }
    isTenantAllowed(context, tenant) {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            if (tenant === this.options.defaultTenant) {
                return true;
            }
            if (this.storage && !(yield this.storage.exists(tenant))) {
                return false;
            }
            return this.options.allowTenant(context, tenant);
        });
    }
    disableTenancyForCurrentScope() {
        this.asyncContext.set(constants_1.MT_SCOPE_KEY, Object.assign(Object.assign({}, this.tenancyScope), { enabled: false }));
        return this;
    }
    setup(options) {
        this.options = Object.assign({}, constants_1.DEFAULT_OPTIONS, options);
        const SettingsDTO = this.options.storageSettingsDto;
        switch (this.options.transport) {
            case core_interface_1.TenantTransport.HTTP:
                this.transport = new http_transport_1.HttpTransport(this.options.headerName, this.options.queryParameterName);
                break;
            default:
                throw new mtenant_error_1.BaseError(`Unknown tenant transport "${this.options.transport}"`);
        }
        if (this.options.storage) {
            switch (typeof this.options.storage) {
                case 'string':
                    switch (this.options.storage.toLowerCase()) {
                        case constants_1.SEQUELIZE_STORAGE:
                            this.storage = new sequelize_storage_1.SequelizeStorage(this.options.storageRepository);
                            break;
                        default:
                            throw new mtenant_error_1.BaseError(`Unrecognized tenant storage type: ${this.options.storage}`);
                    }
                    break;
                default:
                    this.storage = this.options.storage;
            }
        }
        if (this.options.cache) {
            if (!this.storage) {
                throw new mtenant_error_1.BaseError('Cache option must be specified only when storage is set');
            }
            let cache;
            switch (typeof this.options.cache) {
                case 'string':
                    switch (this.options.cache.toLowerCase()) {
                        case constants_1.IOREDIS_CACHE:
                            cache = new ioredis_1.IoRedis(this.options.cacheClient);
                            break;
                        default:
                            throw new mtenant_error_1.BaseError(`Unrecognized tenant storage cache type: ${this.options.cache}`);
                    }
                    break;
                default:
                    cache = this.options.cache;
            }
            this.storage = new cached_storage_1.CachedStorage(this.storage, cache, this.options.cacheOptions);
        }
        for (const entity of this.options.for) {
            (0, entity_1.injectTenancyService)(entity, this);
        }
    }
    get tenancyScope() {
        try {
            return this.asyncContext.get(constants_1.MT_SCOPE_KEY) || this.defaultTenancyScope;
        }
        catch (e) {
            return this.defaultTenancyScope;
        }
    }
    get defaultTenancyScope() {
        return {
            tenant: this.options.defaultTenant,
            enabled: true,
        };
    }
    get tenancyOptions() {
        return this.options;
    }
};
MtenantService = tslib_1.__decorate([
    (0, common_1.Injectable)(),
    tslib_1.__param(0, (0, common_1.Inject)(constants_1.MT_OPTIONS)),
    tslib_1.__metadata("design:paramtypes", [Object, async_context_1.AsyncContext])
], MtenantService);
exports.MtenantService = MtenantService;
