"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.CachedStorage = void 0;
const tslib_1 = require("tslib");
const cache_error_1 = require("./cache/cache.error");
var CachedOp;
(function (CachedOp) {
    CachedOp["EXISTS"] = "exists";
    CachedOp["GET"] = "get";
})(CachedOp || (CachedOp = {}));
class CachedStorage {
    constructor(storage, cache, options) {
        this.storage = storage;
        this.cache = cache;
        this.options = options;
        this.readonly = false;
        this.DEFAULT_OPTIONS = {
            expire: 3600,
            prefix: 'MTENANT_PCACHE/',
        };
        if (!options) {
            this.options = this.DEFAULT_OPTIONS;
        }
        this.options = !options
            ? this.DEFAULT_OPTIONS
            : Object.assign({}, this.DEFAULT_OPTIONS, this.options);
    }
    add(tenant, settings) {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            yield this.purgeCache(tenant);
            return this.storage.add(tenant, settings);
        });
    }
    remove(tenant) {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            yield this.purgeCache(tenant);
            return this.storage.remove(tenant);
        });
    }
    exists(tenant) {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            const key = this.key(tenant, CachedOp.EXISTS);
            if (yield this.cache.has(key)) {
                return this.decode(yield this.cache.get(key));
            }
            const result = yield this.storage.exists(tenant);
            if (!(yield this.cache.set(key, this.encode(result), this.options.expire))) {
                throw new cache_error_1.CacheError(`Unable to store tenant cache for key=${key}`);
            }
            return result;
        });
    }
    updateSettings(tenant, settings) {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            yield this.purgeCache(tenant, CachedOp.GET);
            return this.storage.updateSettings(tenant, settings);
        });
    }
    get(tenant) {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            const key = this.key(tenant, CachedOp.GET);
            if (yield this.cache.has(key)) {
                return this.decode(yield this.cache.get(key));
            }
            const result = yield this.storage.get(tenant);
            if (!(yield this.cache.set(key, this.encode(result), this.options.expire))) {
                throw new cache_error_1.CacheError(`Unable to store tenant cache for key=${key}`);
            }
            return result;
        });
    }
    purgeCache(tenant, op) {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            const ops = op ? [op] : [CachedOp.GET, CachedOp.EXISTS];
            for (const oneOp of ops) {
                const key = this.key(tenant, oneOp);
                if (yield this.cache.has(key)) {
                    if (!(yield this.cache.remove(key))) {
                        throw new cache_error_1.CacheError(`Unable to remove tenant cache for key=${key}`);
                    }
                }
            }
        });
    }
    encode(value) {
        return JSON.stringify(value);
    }
    decode(rawValue) {
        return JSON.parse(rawValue);
    }
    key(tenant, op) {
        return `${this.options.prefix}${tenant}$${op}`;
    }
}
exports.CachedStorage = CachedStorage;
